use num_complex::Complex;
use rayon::prelude::*;

const MAX_ITERATIONS: u16 = 512;

#[derive(Debug, Clone)]
pub struct Mandelbrot {
    iterations: Vec<u16>,
}

enum Colour {
    Purple,
    Violet,
    Blue,
    Green,
    Yellow,
    Orange,
    Red,
    White,
    Black,
}

impl Mandelbrot {
    const IMG_WIDTH: usize = 1280;
    const IMG_HEIGHT: usize = 720;

    pub fn new() -> Self {
        Mandelbrot {
            iterations: vec![0; Self::IMG_WIDTH * Self::IMG_HEIGHT],
        }
    }

    pub fn width() -> usize {
        Self::IMG_WIDTH
    }

    pub fn height() -> usize {
        Self::IMG_HEIGHT
    }

    pub fn generate_image(&mut self) -> Vec<u8> {
        self.escape_time((-2.5, 1.0), (-1.0, 1.0));

        let mut texture_buffer = vec![0; (Self::IMG_WIDTH * Self::IMG_HEIGHT) * 4];
        for i in 0..texture_buffer.len() {
            if i % 4 == 0 {
                let iter_i = i / 4;
                self.assign_colour(i, iter_i, &mut texture_buffer);
            }
        }

        texture_buffer
    }

    pub fn generate_scaled_image(&mut self, x_bounds: (f64, f64), y_bounds: (f64, f64)) -> Vec<u8> {
        let mut texture_buffer = vec![0; (Self::IMG_WIDTH * Self::IMG_HEIGHT) * 4];

        self.escape_time(x_bounds, y_bounds);

        for i in 0..texture_buffer.len() {
            if i % 4 == 0 {
                let iter_i = i / 4;
                self.assign_colour(i, iter_i, &mut texture_buffer);
            }
        }

        texture_buffer
    }

    fn escape_time(&mut self, x_bounds: (f64, f64), y_bounds: (f64, f64)) {
        self.iterations
            .par_iter_mut()
            .enumerate()
            .for_each(|(i, value)| {
                let px = i % Self::IMG_WIDTH;
                let py = i / Self::IMG_WIDTH;

                let c = Self::coords_to_complex(px, py, x_bounds, y_bounds);

                *value = Self::calc_num_iterations(c);
            });
    }

    fn find_colour(iteration: u16) -> Colour {
        if iteration == 1 {
            return Colour::Black;
        }

        match iteration % 7 {
            0 => Colour::Purple,
            1 => Colour::Violet,
            2 => Colour::Blue,
            3 => Colour::Green,
            4 => Colour::Yellow,
            5 => Colour::Orange,
            6 => Colour::Red,
            _ => unreachable!(),
        }
    }

    fn assign_colour(&self, i: usize, iter_i: usize, texture_buffer: &mut [u8]) {
        match Self::find_colour(self.iterations[iter_i]) {
            Colour::White => {
                texture_buffer[i] = 0xFF;
                texture_buffer[i + 1] = 0xFF;
                texture_buffer[i + 2] = 0xFF;
                texture_buffer[i + 3] = 0xFF;
            }
            Colour::Black => {
                texture_buffer[i] = 0x00;
                texture_buffer[i + 1] = 0x00;
                texture_buffer[i + 2] = 0x00;
                texture_buffer[i + 3] = 0xFF;
            }
            Colour::Purple => {
                texture_buffer[i] = 0x94;
                texture_buffer[i + 1] = 0x00;
                texture_buffer[i + 2] = 0xD3;
                texture_buffer[i + 3] = 0xFF;
            }
            Colour::Violet => {
                texture_buffer[i] = 0x4B;
                texture_buffer[i + 1] = 0x00;
                texture_buffer[i + 2] = 0x82;
                texture_buffer[i + 3] = 0xFF;
            }
            Colour::Blue => {
                texture_buffer[i] = 0x00;
                texture_buffer[i + 1] = 0x00;
                texture_buffer[i + 2] = 0xFF;
                texture_buffer[i + 3] = 0xFF;
            }
            Colour::Green => {
                texture_buffer[i] = 0x00;
                texture_buffer[i + 1] = 0xFF;
                texture_buffer[i + 2] = 0x00;
                texture_buffer[i + 3] = 0xFF;
            }
            Colour::Yellow => {
                texture_buffer[i] = 0xFF;
                texture_buffer[i + 1] = 0xFF;
                texture_buffer[i + 2] = 0x00;
                texture_buffer[i + 3] = 0xFF;
            }
            Colour::Orange => {
                texture_buffer[i] = 0xFF;
                texture_buffer[i + 1] = 0x7F;
                texture_buffer[i + 2] = 0x00;
                texture_buffer[i + 3] = 0xFF;
            }
            Colour::Red => {
                texture_buffer[i] = 0xFF;
                texture_buffer[i + 1] = 0x00;
                texture_buffer[i + 2] = 0x00;
                texture_buffer[i + 3] = 0xFF;
            }
        }
    }

    fn calc_num_iterations(c: Complex<f64>) -> u16 {
        let mut z: Complex<f64> = Complex::new(0.0, 0.0);
        let mut num_iterations: u16 = 0;

        for i in 0..MAX_ITERATIONS {
            if z.norm_sqr() > 4.0 {
                num_iterations = i;
                break;
            }

            z = (z * z) + c;
        }

        num_iterations + 1 // account for off by one error
    }

    fn coords_to_complex(
        px: usize,
        py: usize,
        x_bounds: (f64, f64),
        y_bounds: (f64, f64),
    ) -> Complex<f64> {
        Complex::new(
            Self::scale_width(px, x_bounds),
            Self::scale_height(py, y_bounds),
        )
    }

    fn scale_width(px: usize, x_bounds: (f64, f64)) -> f64 {
        // const X_MIN: f64 = -2.5;
        // const X_MAX: f64 = 1.0;

        x_bounds.0 + ((x_bounds.1 - x_bounds.0) * (px as f64 - 0.0)) / Self::IMG_WIDTH as f64 - 0.0
    }

    fn scale_height(py: usize, y_bounds: (f64, f64)) -> f64 {
        // const Y_MIN: f64 = -1.0;
        // const Y_MAX: f64 = 1.0;

        y_bounds.0 + ((y_bounds.1 - y_bounds.0) * (py as f64 - 0.0)) / Self::IMG_HEIGHT as f64 - 0.0
    }
}
